/*
 * BubsyMOD source code - bubsymod.c
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#include <stdio.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "bubsymod.h"
#include "dingoo_buttons.h"
#include "textout.h"
#include "volume.h"
#include "sample_viewer.h"

/* these must be accessable by protracker_sequencer.c */
SDL_Surface *background = NULL, *screen = NULL;

#define SCREEN_FRAMERATE 60
unsigned int next_tick = 0;
static int tick_interval = 1 * 1000 / SCREEN_FRAMERATE;

void syncframe(void)
{
	if (next_tick > SDL_GetTicks())
		SDL_Delay(next_tick - SDL_GetTicks());

	next_tick = SDL_GetTicks() + tick_interval;

	return;
}

int main(int argc, char *argv[])
{
	MODULE *mod = NULL;
	mod_playing = 0;
	
	/* argument 1 = module to play */
	if (argc < 2)
	{
		fprintf(stderr, "Usage: %s <module.mod>\n", argv[0]);
	
		return -1;
	}

	/* initialize SDL */
	if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO) < 0)
	{
		fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
	
		return 1;
	}

	SDL_ShowCursor(SDL_DISABLE);
	SDL_WM_SetCaption("BubsyMOD SDL", NULL);

	/* initialize main surface */
	screen = SDL_SetVideoMode(BUBSYMOD_XRES, BUBSYMOD_YRES, BUBSYMOD_DEPTH, SDL_SWSURFACE);
	if (screen == NULL)
	{
		fprintf(stderr, "Couldn't set video mode: %s\n", SDL_GetError());

		SDL_Quit();

		return 1;
	}

	/* initialize bubsymod */
	if (bubsymodInit() == 0)
	{
		fprintf(stderr, "Couldn't initialize bubsymod!\n"); /* TODO: make bubsymodGetLastError(); */

		SDL_Quit();

		return 1;
	}

	/* load given module */
	mod = bubsymodLoad(argv[1]);
	if (mod == NULL)
	{
		fprintf(stderr, "Couldn't load the module!\n"); /* TODO: make bubsymodGetLastError(); */

		bubsymodClose();
		SDL_Quit();

		return 1;
	}
	
	/* get mixer volume */
	base_volume = getVolume();
	
	/* load background (if any) */
	{
		SDL_Surface *bg_tmp = NULL;
		
		bg_tmp = IMG_Load("background.png");
		if ((bg_tmp == NULL) || (bg_tmp->w != BUBSYMOD_XRES) || (bg_tmp->h != BUBSYMOD_YRES))
		{
			SDL_FreeSurface(bg_tmp);
			bg_tmp = NULL; /* sanity */

			/* make a new fresh surface if it fails */
			background = SDL_CreateRGBSurface(screen->flags, BUBSYMOD_XRES, BUBSYMOD_YRES,
				BUBSYMOD_DEPTH, 0, 0, 0, 0);
			SDL_FillRect(background, NULL, 0x0000);
		}
		else
		{
			/* normalize background surface */
			background = SDL_ConvertSurface(bg_tmp, screen->format, screen->flags);
			SDL_FreeSurface(bg_tmp);
		}
	}

	/* pre-print static strings */
	{
		char pTmp[256];
		sprintf(pTmp, "Song name: %s", mod->head.module_title);
		textOut(background, 4, 10, pTmp, 0xFFFF, 0, 0);
		textOut(background, 4, 1, "BubsyMOD - The simple ProTracker player", 0xFFFF, 0, 0);
	}

	/* start playing the loaded module */
	bubsymodPlay(mod, 0);

	/* main loop - the big things are happening in protracker_sequencer.c */
	{
		int app_running = 1;
		while (app_running)
		{
			SDL_Event event;
			while (SDL_PollEvent(&event))
			{
				/* quit */
				if (event.type == SDL_QUIT)
				{
					app_running = 0;
					break;
				}
				else if (event.type == SDL_KEYDOWN)
				{					
					switch (event.key.keysym.sym)
					{
						/* quit */
						case DINGOO_BUTTON_SELECT:
							app_running = 0;
						break;

						/* volume up */
						case DINGOO_BUTTON_UP:
						{
							if (base_volume <= 95)
							{
								base_volume += 5;
								setVolume(base_volume);
							}
						}
						break;

						/* volume down */
						case DINGOO_BUTTON_DOWN:
						{
							if (base_volume >= 5)
							{
								base_volume -= 5;
								setVolume(base_volume);
							}
						}
						break;

						/* sleep screen */
						case DINGOO_BUTTON_X:
							display_mode = SLEEP_SCREEN;
						break;

						/* sample viewer */
						case DINGOO_BUTTON_Y:
							display_mode = SAMPLE_VIEWER;
						break;

						/* pattern viewer */
						case DINGOO_BUTTON_A:
							display_mode = PATTERN_VIEWER;
						break;
						
						/* quadrascopes */
						case DINGOO_BUTTON_B:
							display_mode = QUADRASCOPES;
						break;

						/* seek forward */
						case DINGOO_BUTTON_R:
							bubsymodJump(mod->cur_order + 1);
						break;
							
						/* seek backwards */
						case DINGOO_BUTTON_L:
							bubsymodJump(mod->cur_order - 1);
						break;
							
						default:
						break;
					}
				}
			}
			
			if (display_mode == QUADRASCOPES)
				drawQuadrascopes(screen, mod);
			
			syncframe();
		}
	}

	/* someone closed the app... let's clean up */
	bubsymodFree(mod);
	bubsymodClose();

	SDL_Quit();

	return 0;
}


