/*
 * BubsyMOD source code - bubsymod.h
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#ifndef __BUBSYMOD_H
#define __BUBSYMOD_H

#define SOUND_FREQUENCY 48000
#define SOUND_BUFFER_SIZE 1024

#define BUBSYMOD_XRES 320
#define BUBSYMOD_YRES 240
#define BUBSYMOD_DEPTH 16

#include <SDL/SDL.h>

typedef struct modnote
{
	int period;
	int sample;
	int command;
	int param;
} modnote_t;

typedef struct
{
	unsigned int total_sample_size;
	int restart_pos;
	int channel_count, order_count, pattern_count;
	int order[128];
	int tempo, ticks;
	char module_title[21];
} MODULE_HEADER;

typedef struct
{
	char text[23];

	unsigned int volbar_hit;

	unsigned int length, offset;
	unsigned int loop_start, loop_length;

	unsigned int finetune;
	int volume;
} MODULE_SAMPLE;

typedef struct
{
	int play_loop;
	int enable_sample_viewer;
	int disable_pattern_viewer, keys_locked;
	int cur_row, cur_order, cur_pattern;
	int cur_bpm, cur_speed;
	int module_loaded;
	MODULE_HEADER head;
	MODULE_SAMPLE samples[31];
	modnote_t *patterns[127];
	char *sample_data;
} MODULE;

enum
{
	PATTERN_VIEWER = 1,
	SAMPLE_VIEWER = 2,
	SLEEP_SCREEN = 3,
	QUADRASCOPES = 4
};

int display_mode;
int mod_playing;

int bubsymodInit();
void bubsymodPlay(MODULE *source, int order);
void bubsymodFree(MODULE *source);
void bubsymodClose();
void bubsymodJump(int order);
MODULE *bubsymodLoad(const char *filename);

#endif
