/*
 * BubsyMOD source code - pattern_viewer.c
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#include <stdio.h>
#include <SDL/SDL.h>
#include "bubsymod.h"
#include "textout.h"
#include "volume.h"

/* pattern viewer defines - kind of hard coded! */
#define PATTERN_ROWS_VISIBLE 25
#define PATTERN_MIDDLE_POS 12
#define PATTERN_START_Y_POS 35
#define PATTERN_START_X_POS 8
#define PATTERN_FONT_HEIGHT 8
#define PATTERN_SPACE 76
#define PATTERN_CURSOR_COL 0x43D3
#define PATTERN_ROW_COL 0xD6BA
#define PATTERN_NOTE_COL 0xFFFF
#define PATTERN_SAMPLE_COL 0x171F
#define PATTERN_EFX_COL 0xFF0F
/* ------------------------------ */

static void get_note_string(int period, char *p_str);
static void get_sample_string(unsigned char p_sample, char *p_str);
static void get_cmd_string(modnote_t note, char *p_str);

void drawPattern(SDL_Surface *screen, SDL_Surface *background, MODULE *source)
{
	char pTmp[256];
	
	if (source->module_loaded && mod_playing)
	{
		/* clear the previous drawings */
		SDL_BlitSurface(background, NULL, screen, NULL);

		/* print module information */
		sprintf(pTmp, "#%03d/%03d - %03dbpm (%02d ticks) - Vol: %03d", source->cur_order, source->head.order_count,
			source->cur_bpm, source->cur_speed, base_volume);
		textOut(screen, 4, 19, pTmp, 0xFFFF, 0, 0);

		{
			int row;
			for (row = 0; row < PATTERN_ROWS_VISIBLE; row++)
			{
				/* clamp the pattern rows */
				if (((source->cur_row+(row-PATTERN_MIDDLE_POS)) >= 0) && ((source->cur_row+(row-PATTERN_MIDDLE_POS)) <= 63))
				{
					/* print row number (to the left) */
					sprintf(pTmp,"%02d",(row-(PATTERN_ROWS_VISIBLE/2))+source->cur_row);
					textOut(screen,0,PATTERN_START_Y_POS+(row*PATTERN_FONT_HEIGHT),pTmp,
						PATTERN_ROW_COL,(row==PATTERN_MIDDLE_POS)?1:0,(row==PATTERN_MIDDLE_POS)?PATTERN_CURSOR_COL:0);

					{
						int column;
						for (column = 0; column < 4; column++)
						{
							/* look up the module note */
							modnote_t note = source->patterns[source->cur_pattern][(((row-(PATTERN_ROWS_VISIBLE/2))+source->cur_row)*4+column)];

							/* print note (if any) */
							get_note_string(note.period, pTmp);
							textOut(screen,PATTERN_START_X_POS+13+(column*PATTERN_SPACE),PATTERN_START_Y_POS+(row*PATTERN_FONT_HEIGHT),pTmp,
								PATTERN_NOTE_COL,(row==PATTERN_MIDDLE_POS)?1:0,(row==PATTERN_MIDDLE_POS)?PATTERN_CURSOR_COL:0);

							/* print sample number (if any) */
							get_sample_string(note.sample, pTmp);
							textOut(screen, PATTERN_START_X_POS+40+(column*PATTERN_SPACE),PATTERN_START_Y_POS+(row*PATTERN_FONT_HEIGHT),pTmp,
								PATTERN_SAMPLE_COL,(row==PATTERN_MIDDLE_POS)?1:0,(row==PATTERN_MIDDLE_POS)?PATTERN_CURSOR_COL:0);

							/* print command and parameter (if any) */
							get_cmd_string(note, pTmp);
							textOut(screen,PATTERN_START_X_POS+59+(column*PATTERN_SPACE),PATTERN_START_Y_POS+(row*PATTERN_FONT_HEIGHT),pTmp,
								PATTERN_EFX_COL,(row==PATTERN_MIDDLE_POS)?1:0,(row==PATTERN_MIDDLE_POS)?PATTERN_CURSOR_COL:0);
						}
					}
				}
			}
		}

		/* flip the backbuffer onto the framebuffer */
		SDL_UpdateRect(screen, 0, 0, 0, 0);
	}
	
	return;
}

static void get_note_string(int period, char *p_str)
{
	static const char name_tab[][4] = 
	{
		"...","C-1","C#1","D-1","D#1","E-1","F-1","F#1","G-1","G#1","A-1","A#1","B-1",
		"C-2","C#2","D-2","D#2","E-2","F-2","F#2","G-2","G#2","A-2","A#2","B-2",
		"C-3","C#3","D-3","D#3","E-3","F-3","F#3","G-3","G#3","A-3","A#3","B-3","???"
	};

	static const int period_tab[37] =
	{
		0,856,808,762,720,678,640,604,570,538,508,480,453,
		428,404,381,360,339,320,302,285,269,254,240,226, 
		214,202,190,180,170,160,151,143,135,127,120,113
	};

	/* look up note by period */
	{
		int i;
		for (i = 0; i < 37; i++)
		{
			if (period == period_tab[i])
				break;
		}

		strcpy(p_str, name_tab[i]);
	}

	return;
}

static void get_sample_string(unsigned char p_sample, char *p_str)
{
	if (p_sample > 0)
		sprintf(p_str, "%02d", p_sample);
	else
		strcpy(p_str, "..");

	return;
}

static void get_cmd_string(modnote_t note, char *p_str)
{
	static const char hextable[16] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

	p_str[0] = hextable[note.command];
	p_str[1] = hextable[note.param >> 0x04];
	p_str[2] = hextable[note.param  & 0x0F];

	/* if no command and parameter, print spaces */
	if ((p_str[0] == '0') && (p_str[1] == '0') && (p_str[2] == '0'))
	{	
		p_str[0] = '.';
		p_str[1] = '.';
		p_str[2] = '.';
	}

	p_str[3] = 0; /* string null termination */

	return;
}
