/*
 * BubsyMOD source code - protracker_sequencer.c
 * This application is open source and can be freely distributed
 * and modified, just leave credits to 8bitbubsy and mukunda.
 *
 * audio_mixer.c written by mukunda and 8bitbubsy
 * bubsymod.c written by 8bitbubsy
 * pattern_viewer.c written by 8bitbubsy
 * protracker_loader.c written by 8bitbubsy
 * protracker_sequencer.c written by mukunda and 8bitbubsy
 * sample_viewer.c written by 8bitbubsy
 * textout.c written by 8bitbubsy
 * volume.c written by 8bitbubsy
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bubsymod.h"
#include "protracker_tables.h"
#include "audio_mixer.h"
#include "pattern_viewer.h"
#include "sample_viewer.h"

/* fractional precision, code taken from ScummVM (/common/frac.h) */
#define FRAC_BITS 14
#define intToFrac(x) (unsigned int)((x) << FRAC_BITS)
#define fracToInt(x) (unsigned int)((x) >> FRAC_BITS)
#define doubleToFrac(x) (unsigned int)((x) * (1L << FRAC_BITS))
/* -------------------------------------------------------------- */

extern SDL_Surface *screen, *background;
char *file_sample_data = NULL;

typedef struct
{
	int seqchannel;

	int rate;
	
	int period, tperiod;
	unsigned int offset, old_offset;

	int pattern_loop_row, pattern_loop_times, pattern_loop_jump;

	int sample;
	int command;
	int param;
	int volume;

	int flags;
	int glismem, vibmem, fxmem;
	
	int ivloop_delay, ivloop_speed;
	unsigned int ivloop_sloopstart, ivloop_sreplen, ivloop_swavestart;
	unsigned int ivloop_soffset;
} mod_channel;

static int mod_speed, mod_bpm;
static int mod_tick, mod_row, mod_pattern, mod_order;

static int mod_samplecounter = 0, mod_samplespertick;

static int pattern_jump_target, pattern_jump_row;
static int pattern_delay, pattern_delay_2;
static int pattern_loop, pattern_loop_row;

static int aperiod;
static int avolume;
static int aflags;

static unsigned int samplespertick_freq = SOUND_FREQUENCY * 5;

static int getperiod(int period, int finetune);
static int getSampleFinetune(int sampleNumber);

static MODULE *source = NULL;
static mod_channel channels[4];

static void set_position(int order, int row);
static void set_tempo(int bpm);
static void set_speed(int speed);

enum
{
	FLAG_NOTE = 1,
	FLAG_SAMPLE = 2,
	FLAG_NEWSAMPLE = 4
};

enum
{
	AFLAG_START = 1,
	AFLAG_DELAY = 2,
};

static void effect_arpeggio(mod_channel *ch);
static void effect_portamento_up(mod_channel *ch);
static void effect_portamento_down(mod_channel *ch);
static void effect_glissando(mod_channel *ch);
static void effect_vibrato(mod_channel *ch);
static void effect_glissando_vslide(mod_channel *ch);
static void effect_vibrato_vslide(mod_channel *ch);
static void effect_tremolo(mod_channel *ch);
static void effect_setpanning(mod_channel *ch);
static void effect_sample_offset(mod_channel *ch);
static void effect_volume_slide(mod_channel *ch);
static void effect_position_jump(mod_channel *ch);
static void effect_set_volume(mod_channel *ch);
static void effect_pattern_break(mod_channel *ch);
static void effect_extended(mod_channel *ch);
static void effect_tempo(mod_channel *ch);
static void effecte_setfilter(mod_channel *ch);
static void effecte_fineportaup(mod_channel *ch);
static void effecte_fineportadown(mod_channel *ch);
static void effecte_glissandoctrl(mod_channel *ch);
static void effecte_vibratoctrl(mod_channel *ch);
static void effecte_setfinetune(mod_channel *ch);
static void effecte_patternloop(mod_channel *ch);
static void effecte_tremoloctrl(mod_channel *ch);
static void effecte_setpanning(mod_channel *ch);
static void effecte_retrignote(mod_channel *ch);
static void effecte_finevolup(mod_channel *ch);
static void effecte_finevoldown(mod_channel *ch);
static void effecte_notecut(mod_channel *ch);
static void effecte_notedelay(mod_channel *ch);
static void effecte_patterndelay(mod_channel *ch);
static void effecte_invertloop(mod_channel *ch);

typedef void (*effect_routine)(mod_channel *);

static effect_routine effect_routines[] =
{
	effect_arpeggio,
	effect_portamento_up,
	effect_portamento_down,
	effect_glissando,
	effect_vibrato,
	effect_glissando_vslide,
	effect_vibrato_vslide,
	effect_tremolo,
	effect_setpanning,
	effect_sample_offset,
	effect_volume_slide,
	effect_position_jump,
	effect_set_volume,
	effect_pattern_break,
	effect_extended,
	effect_tempo
};
static effect_routine effecte_routines[] =
{
	effecte_setfilter,
	effecte_fineportaup,
	effecte_fineportadown,
	effecte_glissandoctrl,
	effecte_vibratoctrl,
	effecte_setfinetune,
	effecte_patternloop,
	effecte_tremoloctrl,
	effecte_setpanning,
	effecte_retrignote,
	effecte_finevolup,
	effecte_finevoldown,
	effecte_notecut,
	effecte_notedelay,
	effecte_patterndelay,
	effecte_invertloop
};

static void UpdateInvertLoop(mod_channel *ch)
{
	char* smp;
	unsigned int tmp_swavestart;
	
	if (ch->ivloop_speed > 0)
	{
		ch->ivloop_delay += ivloop_table[ch->ivloop_speed];
		if ((ch->ivloop_delay & 128))
		{
			ch->ivloop_delay = 0;

			tmp_swavestart = ch->ivloop_swavestart + 1;
			if (tmp_swavestart >= (ch->ivloop_sloopstart + ch->ivloop_sreplen))
				tmp_swavestart = ch->ivloop_sloopstart;
				
			ch->ivloop_swavestart = tmp_swavestart;

			smp = &source->sample_data[ch->ivloop_soffset + tmp_swavestart];
			*smp = ~*smp;
		}
	}
	
	return;
}

static void effecte_setfilter(mod_channel *ch)
{
	return;
}
static void effecte_fineportaup(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		ch->period -= (ch->param & 0x0F);
		if (ch->period < 113) ch->period = 113;
		aperiod = ch->period;
	}
	
	return;
}
static void effecte_fineportadown(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		ch->period += (ch->param & 0x0F);
		if (ch->period > 856) ch->period = 856;
		aperiod = ch->period;
	}
	
	return;
}
static void effecte_glissandoctrl(mod_channel *ch)
{
	return;
}
static void effecte_vibratoctrl(mod_channel *ch)
{
	return;
}
static void effecte_setfinetune(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (ch->sample > 0)
			source->samples[ch->sample - 1].finetune = ch->param;
	}
	
	return;
}
static void effecte_patternloop(mod_channel *ch)
{
	/* code from ptplay.c - 
	 * thanks for figuring out how this is
	 * supposed to be done!
	 */

	if (mod_tick == 0)
	{
		int pattern_loop_parm = ch->param & 0x0F;
		if (ch->pattern_loop_jump == 0)
		{
			if (pattern_loop_parm == 0)
			{
				ch->pattern_loop_row = mod_row;
				ch->pattern_loop_jump = 1;

				return;
			}
		}

		if (pattern_loop_parm != 0)
		{
			if (ch->pattern_loop_jump != 2)
				ch->pattern_loop_times = pattern_loop_parm;

			pattern_loop = 1;

			ch->pattern_loop_jump = 2;
			if (ch->pattern_loop_times <= 0)
			{
				ch->pattern_loop_times = 0;
				ch->pattern_loop_jump = 0;
				pattern_loop = 0;
			}
			else
			{
				pattern_loop = 1;
				pattern_loop_row = ch->pattern_loop_row;

				ch->pattern_loop_times--;
			}
		}
	}
	
	return;
}
static void effecte_tremoloctrl(mod_channel *ch)
{
	return;
}
static void effecte_setpanning(mod_channel *ch)
{
	return;
}
static void effecte_retrignote(mod_channel *ch)
{
	int what_tick = ch->param & 0x0F;
	
	if (what_tick > 0)
	{
		if ((mod_tick % what_tick) == 0)
			aflags |= AFLAG_START;
	}
	
	return;
}
static void effecte_finevolup(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		ch->volume += (ch->param & 0x0F);
		if (ch->volume > 64) ch->volume = 64;
		avolume = ch->volume;
	}
	
	return;
}
static void effecte_finevoldown(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		ch->volume -= (ch->param & 0x0F);
		if (ch->volume < 0) ch->volume = 0;
		avolume = ch->volume;
	}
	
	return;
}
static void effecte_notecut(mod_channel *ch)
{
	if (mod_tick == (ch->param & 0x0F))
		ch->volume = avolume = 0;
		
	return;
}
static void effecte_notedelay(mod_channel *ch)
{
	unsigned char x = ch->param & 0x0F;

	if (mod_tick == 0)
		ch->fxmem = aflags;
		
	if (mod_tick < x)
		aflags |= AFLAG_DELAY;
	else if (mod_tick == x)
		aflags = ch->fxmem;
	
	return;
}
static void effecte_patterndelay(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (pattern_delay_2 == 0)
			pattern_delay = (ch->param & 0x0F) + 1;
	}
	
	return;
}
static void effecte_invertloop(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		ch->ivloop_speed = ch->param & 0x0F;
		UpdateInvertLoop(ch);
	}
	
	return;
}
static void do_glissando(mod_channel *ch)
{
	if (mod_tick > 0)
	{
		if (ch->period < ch->tperiod)
		{
			ch->period += ch->glismem;
			if (ch->period > ch->tperiod) ch->period = ch->tperiod;
		}
		else
		{
			ch->period -= ch->glismem;
			if (ch->period < ch->tperiod) ch->period = ch->tperiod;
		}
		aperiod = ch->period;
	}
	
	return;
}
static void do_vibrato(mod_channel *ch)
{
	if (ch->fxmem < 32)
		aperiod += (sin_table[ch->fxmem & 0x1F] * (ch->vibmem & 0x0F)) >> 7;
	else
		aperiod -= (sin_table[ch->fxmem & 0x1F] * (ch->vibmem & 0x0F)) >> 7;

	if (mod_tick > 0)
	{
		ch->fxmem += (ch->vibmem >> 4);
		ch->fxmem &= 0x3F;
	}
	
	return;
}
static void effect_arpeggio(mod_channel *ch)
{
	unsigned char tick3 = mod_tick % 3;
	
	if (tick3 == 1)
		aperiod = (aperiod * arp_table[ch->param >> 4]) >> 15;
	else if (tick3 == 2)
		aperiod = (aperiod * arp_table[ch->param & 0x0F]) >> 15;
	
	return;
}
static void effect_portamento_up(mod_channel *ch)
{
	if (mod_tick > 0)
	{
		ch->period -= ch->param;
		if (ch->period < 113) ch->period = 113;
		aperiod = ch->period;
	}
	
	return;
}
static void effect_portamento_down(mod_channel *ch)
{
	if (mod_tick > 0)
	{
		ch->period += ch->param;
		if (ch->period > 856) ch->period = 856;
		aperiod = ch->period;
	}
	
	return;
}
static void effect_glissando(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (ch->param > 0)
			ch->glismem = ch->param;
	}
	do_glissando(ch);
	
	return;
}
static void effect_vibrato(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (ch->param & 0xF0)
			ch->vibmem = (ch->vibmem & 0x0F) | (ch->param & 0xF0);
		if (ch->param & 0x0F)
			ch->vibmem = (ch->vibmem & 0xF0) | (ch->param & 0x0F);
	}
	do_vibrato(ch);
	
	return;
}
static void effect_glissando_vslide(mod_channel *ch)
{
	do_glissando(ch);
	effect_volume_slide(ch);
	
	return;
}
static void effect_vibrato_vslide(mod_channel *ch)
{
	do_vibrato(ch);
	effect_volume_slide(ch);
	
	return;
}
static void effect_tremolo(mod_channel *ch)
{
	if (mod_tick > 0)
	{
		ch->fxmem += (ch->param >> 4);
		ch->fxmem &= 0x3F;
	}

	if (ch->fxmem < 32)
	{
		avolume += (sin_table[ch->fxmem & 0x1F] * (ch->param & 0x0F)) >> 6;
		if (avolume > 64) avolume = 64;
	}
	else
	{
		avolume -= (sin_table[ch->fxmem & 0x1F] * (ch->param & 0x0F)) >> 6;
		if (avolume < 0) avolume = 0;
	}

	return;
}
static void effect_setpanning(mod_channel *ch)
{
	return;
}
static void effect_sample_offset(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (ch->param > 0)
			ch->offset = ch->param << 8;
	}
	
	return;
}
static void effect_volume_slide(mod_channel *ch)
{
	if (mod_tick > 0)
	{
		if (ch->param & 0xF0)
		{
			avolume += (ch->param >> 4);
			if (avolume > 64) avolume = 64;
			ch->volume = avolume;
		}
		else
		{
			avolume -= (ch->param & 0x0F);
			if (avolume < 0) avolume = 0;
			ch->volume = avolume;
		}
	}
	
	return;
}
static void effect_position_jump(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (pattern_jump_row == 0)
			pattern_jump_row = 1;

		pattern_jump_target = ch->param;
	}
	
	return;
}
static void effect_set_volume(mod_channel *ch)
{
	if (mod_tick == 0)
		ch->volume = ch->param;
	
	return;
}
static void effect_pattern_break(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (pattern_jump_row == 0)
			pattern_jump_target = mod_order + 1;

		if (source->head.order_count == 1)
			pattern_jump_target = 0;

		pattern_jump_row = ((((ch->param & 0xF0) >> 4) * 10) + (ch->param & 0x0F)) + 1;
		if (pattern_jump_row > 64)
			pattern_jump_row = 1;
	}
	
	return;
}
static void effect_extended(mod_channel *ch)
{
	effecte_routines[ch->param >> 4](ch);
	return;
}
static void effect_tempo(mod_channel *ch)
{
	if (mod_tick == 0)
	{
		if (ch->param > 32)
			set_tempo(ch->param);
		else
			set_speed(ch->param);
	}
	
	return;
}

static void update_effect(mod_channel *ch)
{
	UpdateInvertLoop(ch);
	effect_routines[ch->command](ch);
	
	return;
}

static void read_note(mod_channel *ch)
{
	if (mod_playing)
	{
		modnote_t *note = &source->patterns[mod_pattern][mod_row * 4 + ch->seqchannel];
		if (note->sample > 0)
		{
			if (ch->sample != note->sample)
				ch->flags |= FLAG_NEWSAMPLE;
				
			ch->sample = note->sample;
			ch->flags |= FLAG_SAMPLE;
			
			/* update invert loop variables */
			{
				MODULE_SAMPLE *s = &source->samples[ch->sample - 1];

				if (s->loop_length > 0)
				{
					ch->ivloop_swavestart = s->loop_start;
					ch->ivloop_soffset    = s->offset;
					ch->ivloop_sreplen    = s->loop_length;
					ch->ivloop_sloopstart = s->loop_start;
				}
				else
				{
					ch->ivloop_swavestart = 0;
					ch->ivloop_soffset    = s->offset;
					ch->ivloop_sreplen    = 0;
					ch->ivloop_sloopstart = 0;
				}
			}
		}

		ch->command = note->command;
		ch->param = note->param;

		if (note->period > 0)
		{
			ch->tperiod = getperiod(note->period, getSampleFinetune(ch->sample));
			ch->flags |= FLAG_NOTE;
		}
	}
	
	return;
}

static void update_channel(mod_channel *ch)
{
	aflags = 0;
	if (mod_tick == 0)
	{
		if (pattern_delay_2 == 0)
			read_note(ch);

		/* period found (e.g. > 0) */
		if ((ch->flags & FLAG_NOTE))
		{
			ch->flags &= ~FLAG_NOTE;

			/* don't use temp period and don't reset offset on 3xx or 5xx */
			if ((ch->command != 3) && (ch->command != 5))
			{
				ch->period = ch->tperiod;
				
				if (ch->command != 9)
					ch->offset = 0;
				
				aflags |= AFLAG_START;
			}

			ch->fxmem = 0;
		}

		/* if a sample number was found */
		if ((ch->flags & FLAG_SAMPLE))
		{
			ch->flags &= ~FLAG_SAMPLE;
			
			if (ch->sample > 0)
			{
				ch->volume = source->samples[ch->sample - 1].volume;
				
				/* new sample number found, differs from earlier one */
				if (ch->flags & FLAG_NEWSAMPLE)
				{
					MODULE_SAMPLE *s = &source->samples[ch->sample - 1];
					
					if (ch->command != 9)
						ch->offset = 0;
						
					/* FIXED: use the change function here, so we don't reset the channel read position */
					audioChangeSrc(ch->seqchannel, source->sample_data + s->offset, s->length, s->loop_start, s->loop_length);
						
					ch->flags &= ~FLAG_NEWSAMPLE;
				}
			}
		}
	}

	aperiod = ch->period;
	avolume = ch->volume;

	update_effect(ch);

	/* if NOT note delay, proceed note */
	if (!(aflags & AFLAG_DELAY))
	{
		/* if the trigger sample flag is on */
		if ((aflags & AFLAG_START))
		{
			if (ch->sample > 0)
			{
				MODULE_SAMPLE *s = &source->samples[ch->sample - 1];
				audioSetSource(ch->seqchannel, source->sample_data + s->offset, s->length, s->loop_start, s->loop_length);
				
				if (ch->offset > 0)
					audioSetOffset(ch->seqchannel, ch->offset);

				if (display_mode == SAMPLE_VIEWER)
				{
					s->volbar_hit = ch->volume >> 1;
					if (s->volbar_hit > 27)
						s->volbar_hit = 27;
				}
			}
		}
		
		audioSetVolume(ch->seqchannel, ch->volume);
		
		/* calculate increase rate from period */
		if (aperiod > 0)
		{
			ch->rate = doubleToFrac((double)(7093789.2 / (aperiod * 2.0))) / SOUND_FREQUENCY;
			audioSetRate(ch->seqchannel, ch->rate);
		}
	}
	
	return;
}

static void process_tick()
{
	if (mod_tick == 0)
	{
		if (pattern_delay)
		{
			pattern_delay_2 = pattern_delay;
			pattern_delay = 0;
		}

		if (pattern_delay_2 > 0)
			pattern_delay_2--;
	}

	if (mod_playing)
	{
		update_channel(channels);
		update_channel(channels + 1);
		update_channel(channels + 2);
		update_channel(channels + 3);
	}

	switch (display_mode)
	{
		case PATTERN_VIEWER:
			drawPattern(screen, background, source); /* we draw/update the pattern viewer each tick */
		break;

		case SAMPLE_VIEWER:
		{
			drawSampleViewer(screen, source);
			
			{
				int i;
				for (i = 0; i < 31; i++)
				{
					MODULE_SAMPLE *s = &source->samples[i];
					if (s->volbar_hit > 0)
						s->volbar_hit--;
				}
			}
		}

		case QUADRASCOPES:
		case SLEEP_SCREEN:
		default:
			break;
	}

	if (++mod_tick >= mod_speed)
	{
		mod_tick = 0;

		if (pattern_loop == 1)
		{
			set_position(mod_order, pattern_loop_row);
			pattern_loop = 0;
		}
		else if (pattern_delay_2 == 0)
		{
			if (pattern_jump_row)
			{
				set_position(pattern_jump_target, pattern_jump_row - 1);
			}
			else
			{
				if (++mod_row >= 64)
					set_position(mod_order + 1, 0);
			}

			source->cur_row = mod_row;
		}
	}

	return;
}

int bubsymodPulse(int samples)
{
	if (mod_playing)
	{
		if (mod_samplecounter == 0)
		{
			process_tick();
			mod_samplecounter += mod_samplespertick;
		}

		mod_samplecounter -= samples;
		if (mod_samplecounter < 0)
		{
			samples += mod_samplecounter;
			mod_samplecounter = 0;
		}
	}

	return samples;
}

static void set_position(int order, int row)
{
	if (order >= source->head.order_count)
	{
		if (source->play_loop == 1)
			order = source->head.restart_pos;
		else
		{
			order = 0;
			row = 0;
		}
	}
		
	source->cur_order = mod_order = order;
	source->cur_row = mod_row = row;
	source->cur_pattern = mod_pattern = source->head.order[order];
	
	pattern_jump_row = 0;
	pattern_jump_target = 0;

	pattern_delay = 0;
	pattern_delay_2 = 0;

	pattern_loop = 0;
	pattern_loop_row = 0;

	mod_tick = 0;
	
	return;
}

void bubsymodJump(int order)
{
	set_position(order, 0);
		
	return;
}

static void set_source(MODULE *src)
{
	source = src;
	return;
}

static void set_speed(int speed)
{
	source->cur_speed = mod_speed = speed;
	
	return;
}

static void set_tempo(int bpm)
{
	source->cur_bpm = mod_bpm = bpm;
	mod_samplespertick = samplespertick_freq / (bpm << 1);
	
	return;
}

int bubsymodInit()
{
	display_mode = PATTERN_VIEWER;
	return audioStart();
}

void bubsymodPlay(MODULE *mod, int order)
{
	audioCutAllVoices();

	channels[0].volume = 64;
	channels[0].seqchannel = 0;
	channels[1].volume = 64;
	channels[1].seqchannel = 1;
	channels[2].volume = 64;
	channels[2].seqchannel = 2;
	channels[3].volume = 64;
	channels[3].seqchannel = 3;
	
	set_source(mod);
	set_speed(6);
	set_tempo(125);
	set_position(order, 0);

	source->cur_row = mod_row = 0;

	pattern_delay = 0;
	pattern_delay_2 = 0;

	pattern_loop = 0;
	pattern_loop_row = 0;

	mod_tick = 0;
	mod_playing = 1;
	
	return;
}

void bubsymodFree(MODULE* source)
{
	if (source != NULL)
	{
		mod_playing = 0;
		
		SDL_Delay(10); // make sure we stop all pattern readings

		{
			int i;
			for (i = 0; i < source->head.pattern_count; i++)
				free(source->patterns[i]);
		}
		
		if (file_sample_data != NULL)
			free(file_sample_data);

		free(source);
	}
	
	return;
}

void bubsymodClose()
{
	audioStop();
	return;
}

static int getperiod(int period, int finetune)
{
	int i;
	for (i = 0; i < 36; i++)
	{
		if (tab_tunes[0][i] == period)
			break;
	}

	return tab_tunes[finetune][i];
}

static int getSampleFinetune(int sampleNumber)
{
	if ((sampleNumber >= 1) && (sampleNumber <= 31))
		return source->samples[sampleNumber - 1].finetune;

	return 0;
}
